# General packages
library(dplyr)
library(tidyr)
library(furrr)

# Geo packages
library(terra)
library(CoordinateCleaner)
library(sf)
library(geos)

# Stat packages
library(MASS)
library(dismo)
library(spatialEco)
library(blockCV)
library(caret)

source("R/utils.R")

load("data/r_objects/range_maps.RData")
load("data/r_objects/occs_final.RData")

# ---------------------------------------------------------------------------#
# Prepare Geodata                                                         ####
# ---------------------------------------------------------------------------#
proj_string = "+proj=tcea +lon_0=-58.0704167 +datum=WGS84 +units=m +no_defs" 

target_species = unique(occs_final$species)

sa_polygon = rnaturalearth::ne_countries() %>% 
  dplyr::filter(continent == "South America") %>% 
  sf::st_union() 

raster_filepaths = list.files("~/symobio-modeling/data/geospatial/raster/", full.names = T) %>% 
  stringr::str_sort(numeric = T) 

sa_polygon = st_transform(sa_polygon, proj_string)
occs_final = st_transform(occs_final, proj_string)
range_maps = st_transform(range_maps, proj_string)
raster_data = terra::rast(raster_filepaths) %>% 
  terra::project(proj_string) 

# ---------------------------------------------------------------------------#
# Sample random absences                                                  ####
#                                                                            #
# Absences are sampled at random                                             #
# (1) within the defined sampling area (same number as presences)            #
# (2) outside the defined sampling area (100 records)                        #
# ---------------------------------------------------------------------------#
species_processed = list.files("data/r_objects/pa_sampling_random_extended/", pattern = "RData") %>% 
  stringr::str_remove(".RData")

for(species in setdiff(target_species, species_processed)){
  print(species)
  # Prepare species occurrence data #####
  occs_spec = dplyr::filter(occs_final, species == !!species)
  if(nrow(occs_spec) <= 1){
    warning("Skipping species ", species, ": Can't process species with a single presence record." )
  }
  
  occs_multipoint = occs_spec %>% 
    summarise(geometry = st_combine(geometry))
  
  range_spec = dplyr::filter(range_maps, name_matched == !!species) %>% 
    rename(species = name_matched)
  
  # Define core sampling region
  # 1. Union all points from range polygon and occurrences records
  # 2. Build concave polygon from unioned points
  # 3. Buffer by 50 km to extend sampling region into new environments
  # 4. Crop by South America outline
  # 5. Crop by buffered presence points
  sampling_region = st_cast(range_spec, "MULTIPOINT") %>% 
    st_union(occs_multipoint) %>% 
    geos::geos_concave_hull(ratio = 0.3) %>% # Concave hull in sf requires geos 3.11
    st_as_sf() %>% 
    st_set_crs(proj_string) %>% 
    st_buffer(dist = 50000) %>% 
    st_intersection(sa_polygon) %>% 
    st_difference(st_buffer(occs_multipoint, dist = 50000))
  
  # Sample balanced number of pseudo absences inside core region
  abs_spec_list = list()
  
  samples_required = nrow(occs_spec)
  while(samples_required != 0){
    sample_points = sf::st_sample(sampling_region, size = samples_required)
    env_sample = terra::extract(raster_data, terra::vect(sample_points), ID = FALSE)
    
    sample_abs = st_sf(env_sample, geometry = sample_points) %>% 
      drop_na() %>% 
      dplyr::mutate(
        record_type = "absence_core"
      )
    
    abs_spec_list[[length(abs_spec_list)+1]] = sample_abs
    
    samples_required = samples_required - nrow(sample_abs) # Sometimes there are no env data for sample points, so keep sampling
  }
  
  # Sample 100 pseudo absences outside core region
  background_region = st_difference(sa_polygon, sampling_region)
  samples_required = 100
  while(samples_required != 0){
    sample_points = sf::st_sample(background_region, size = samples_required)
    env_sample = terra::extract(raster_data, terra::vect(sample_points), ID = FALSE)
    
    sample_abs = st_sf(env_sample, geometry = sample_points) %>% 
      drop_na() %>% 
      dplyr::mutate(
        record_type = "absence_background"
      )
    
    abs_spec_list[[length(abs_spec_list)+1]] = sample_abs
    
    samples_required = samples_required - nrow(sample_abs) # Sometimes there are no env data for sample points, so keep sampling
  }
  
  # Consolidate absences
  abs_spec = bind_rows(abs_spec_list)
  abs_spec = abs_spec %>% 
    bind_cols(
      st_coordinates(st_transform(abs_spec, "EPSG:4326")) 
    ) %>% 
    dplyr::rename(
      longitude = X,
      latitude = Y
    ) %>% 
    dplyr::mutate(
      species = !!species,
      present = 0
    ) 
  
  # Create presence-absence dataframe
  pa_spec = occs_spec %>% 
    dplyr::mutate(
      present = 1,
      record_type = "present"
    ) %>% 
    bind_rows(abs_spec)
  
  # ref = st_bbox(sampling_region)[c(1,3,2,4)]
  # ggplot() +
  #   ggtitle(species) +
  #   geom_sf(data = st_crop(sa_polygon, ref), fill = "#756445") +
  #   geom_sf(data = range_spec, alpha = 0, color = "black") +
  #   geom_sf(data = pa_spec, aes(color = as.factor(present)), size = 0.7) +
  #   scale_color_discrete(name = "Presence") +
  #   theme_minimal()
  # 
  # try({
  #   ggsave(paste0(species, ".pdf"), path = "plots/pa_sampling_random", device = "pdf", scale = 2)
  # })
  
  # Define cross-validation folds
  folds = tryCatch({
    spatial_folds = suppressMessages(
      blockCV::cv_spatial(
        pa_spec,
        column = "present",
        k = 5,
        progress = F, plot = F, report = F
      )
    )
    
    spatial_folds$folds_ids
  }, error = function(e){
    NA
  })
  
  pa_spec$fold_eval = folds
  
  save(pa_spec, file = paste0("data/r_objects/pa_sampling_random_extended/", species, ".RData"))
}

# Combine presence-absence data
files = list.files("data/r_objects/pa_sampling_random_extended/", full.names = T)
model_data = lapply(files, function(f){load(f); return(pa_spec)}) %>% 
  bind_rows() %>% 
  group_by(species, present) %>%
  mutate(weight = 1/n()) %>%
  ungroup()

save(model_data, file = "data/r_objects/model_data_random_abs_extended.RData")
