library(dplyr)
library(tidyr)
library(sf)
library(caret)
library(cito)
library(pROC)

source("R/utils.R")

load("data/r_objects/model_data.RData")

# ----------------------------------------------------------------------#
# Run training                                                       ####
# ----------------------------------------------------------------------#
species_processed = list.files("data/r_objects/ssdm_results/performance/", pattern = "RData") %>% 
  stringr::str_remove(".RData")

data_split = model_data %>% 
  dplyr::filter(!is.na(fold_eval) & !species %in% species_processed) %>% 
  dplyr::group_by(species) %>% 
  dplyr::group_split()

for(pa_spec in data_split){
  species = pa_spec$species[1]
  print(species)

  # Define empty result for performance eval
  na_performance = list(AUC = NA_real_, Accuracy = NA_real_, Kappa = NA_real_, 
                        Precision = NA_real_, Recall = NA_real_, F1 = NA_real_, 
                        TP = NA_real_, FP = NA_real_, TN = NA_real_, FN = NA_real_)
  
  # Create factor presence column 
  pa_spec$present_fct = factor(pa_spec$present, levels = c("0", "1"), labels = c("A", "P"))
  
  # Outer CV loop (for averaging performance metrics)
  performance_cv = lapply(1:5, function(fold){
    print(paste("Fold", fold))
    
    ## Preparations #####
    data_test = dplyr::filter(pa_spec, fold_eval == fold)
    data_train = dplyr::filter(pa_spec, fold_eval != fold)
    
    # Create inner CV folds for model training
    cv_train = blockCV::cv_spatial(
      data_train,
      column = "present",
      k = 5,
      progress = F, plot = F, report = F
    )
    data_train$fold_train = cv_train$folds_ids
    
    # Drop geometry
    data_train$geometry = NULL
    data_test$geometry = NULL
    
    # Define caret training routine 
    index_train = lapply(unique(sort(data_train$fold_train)), function(x){
      return(which(data_train$fold_train != x))
    })
    
    train_ctrl = caret::trainControl(
      search = "random",
      classProbs = TRUE, 
      index = index_train,
      summaryFunction = caret::twoClassSummary, 
      savePredictions = "final"
    )
    
    # Define predictors
    predictors = c("bio6", "bio17", "cmi", "rsds", "igfc", "dtfw", "igsw", "roughness")
    
    ## Random Forest #####
    rf_performance = tryCatch({
      # Fit model
      rf_fit = caret::train(
        x = data_train[, predictors],
        y = data_train$present_fct,
        method = "rf",
        trControl = train_ctrl,
        tuneLength = 8,
        verbose = F
      )
      
      evaluate_model(rf_fit, data_test)
    }, error = function(e){
      na_performance
    })
    
    ## Gradient Boosted Machine ####
    gbm_performance = tryCatch({
      gbm_fit = train(
        x = data_train[, predictors],
        y = data_train$present_fct,
        method = "gbm",
        trControl = train_ctrl,
        tuneLength = 8,
        verbose = F
      )
      evaluate_model(gbm_fit, data_test)
    }, error = function(e){
      na_performance
    })
    
    ## Generalized Additive Model ####
    gam_performance = tryCatch({
      gam_fit = train(
        x = data_train[, predictors],
        y = data_train$present_fct,
        method = "gamSpline",
        tuneLength = 8,
        trControl = train_ctrl
      )
      evaluate_model(gam_fit, data_test)
    }, error = function(e){
      na_performance
    })
    
    ## Neural Network ####
    formula = as.formula(paste0("present ~ ", paste(predictors, collapse = '+')))
    nn_performance = tryCatch({
      nn_fit = dnn(
        formula,
        data = data_train,
        hidden = c(200L, 200L, 200L),
        loss = "binomial",
        epochs = 200L, 
        burnin = 200L,
        early_stopping = 30,
        lr = 0.001,   
        batchsize = min(ceiling(nrow(data_train)/10), 64),
        dropout = 0.25,
        optimizer = config_optimizer("adam"),
        validation = 0.2,
        device = "cuda",
        verbose = F,
        plot = F
      )
      
      evaluate_model(nn_fit, data_test)  
    }, error = function(e){
      na_performance
    })
    
    # Summarize results
    performance_summary = tibble(
      species = !!species,
      obs = nrow(data_train),
      fold_eval = fold,
      model = c("RF", "XGB", "GAM", "NN"),
      AUC = c(rf_performance$AUC, xgb_performance$AUC, gam_performance$AUC, nn_performance$AUC),
      Accuracy = c(rf_performance$Accuracy, xgb_performance$Accuracy, gam_performance$Accuracy, nn_performance$Accuracy),
      Kappa = c(rf_performance$Kappa, xgb_performance$Kappa, gam_performance$Kappa, nn_performance$Kappa),
      Precision = c(rf_performance$Precision, xgb_performance$Precision, gam_performance$Precision, nn_performance$Precision),
      Recall = c(rf_performance$Recall, xgb_performance$Recall, gam_performance$Recall, nn_performance$Recall),
      F1 = c(rf_performance$F1, xgb_performance$F1, gam_performance$F1, nn_performance$F1),
      TP = c(rf_performance$TP, xgb_performance$TP, gam_performance$TP, nn_performance$TP),
      FP = c(rf_performance$FP, xgb_performance$FP, gam_performance$FP, nn_performance$FP),
      TN = c(rf_performance$TN, xgb_performance$TN, gam_performance$TN, nn_performance$TN),
      FN = c(rf_performance$FN, xgb_performance$FN, gam_performance$FN, nn_performance$FN)
    ) %>% 
      tidyr::pivot_longer(-any_of(c("species", "obs", "fold_eval", "model")), names_to = "metric", values_to = "value") 
    
    return(performance_summary)
  })
  
  # Combine and save evaluation results
  performance_spec = bind_rows(performance_cv) %>% 
    dplyr::arrange(fold_eval, model, metric)
  
  save(performance_spec, file = paste0("data/r_objects/ssdm_results/performance/", species, ".RData"))
}

# Combine results  
files = list.files("data/r_objects/ssdm_results/performance/", full.names = T, pattern = ".RData")
ssdm_results = lapply(files, function(f){load(f); return(performance_spec)}) %>% 
  bind_rows() 

save(ssdm_results, file = "data/r_objects/ssdm_performance.RData")

# ----------------------------------------------------------------------#
# Train full models                                                  ####
# ----------------------------------------------------------------------#
data_split = model_data %>% 
  dplyr::filter(!is.na(fold_eval)) %>% 
  dplyr::mutate(present_fct = factor(present, levels = c("0", "1"), labels = c("A", "P"))) %>%
  dplyr::group_by(species) %>% 
  dplyr::group_split()

for(pa_spec in data_split){
  species = pa_spec$species[1]
  print(species)
  
  # Create inner CV folds for model training
  cv_train = blockCV::cv_spatial(
    pa_spec,
    column = "present",
    k = 5,
    progress = F, plot = F, report = F
  )
  pa_spec$fold_train = cv_train$folds_ids
  
  # Drop geometry
  pa_spec$geometry = NULL
  
  # Define caret training routine 
  index_train = lapply(unique(sort(pa_spec$fold_train)), function(x){
    return(which(pa_spec$fold_train != x))
  })
  
  train_ctrl = caret::trainControl(
    search = "random",
    classProbs = TRUE, 
    index = index_train,
    summaryFunction = caret::twoClassSummary, 
    savePredictions = "final",
  )

  # Define predictors
  predictors = c("bio6", "bio17", "cmi", "rsds", "igfc", "dtfw", "igsw", "roughness")
  
  # Fit models
  try({
    # Fit model
    rf_fit = caret::train(
      x = pa_spec[, predictors],
      y = pa_spec$present_fct,
      method = "rf",
      metric = "Kappa",
      trControl = train_ctrl,
      tuneLength = 8,
      verbose = F
    )

    save(rf_fit, file = paste0("data/r_objects/ssdm_results/full_models/", species, "_rf_fit.RData"))
  })
  
  try({
    xgb_fit = train(
      x = pa_spec[, predictors],
      y = pa_spec$present_fct,
      method = "xgbTree",
      metric = "Kappa",
      trControl = train_ctrl,
      tuneLength = 8,
      verbose = F
    )
    save(xgb_fit, file = paste0("data/r_objects/ssdm_results/full_models/", species, "_xgb_fit.RData"))
  })
  
  try({
    gam_fit = train(
      x = pa_spec[, predictors],
      y = pa_spec$present_fct,
      method = "gamSpline",
      metric = "Kappa",
      trControl = train_ctrl,
      tuneLength = 8,
    )
    save(gam_fit, file = paste0("data/r_objects/ssdm_results/full_models/", species, "_gam_fit.RData"))
  })
  
  try({
    formula = as.formula(paste0("present ~ ", paste(predictors, collapse = '+')))
    nn_fit = dnn(
      formula,
      data = pa_spec,
      hidden = c(200L, 200L, 200L),
      loss = "binomial",
      epochs = 200L, 
      burnin = 200L,
      early_stopping = 30,
      lr = 0.001,   
      batchsize = min(ceiling(nrow(pa_spec)/10), 64),
      dropout = 0.25,
      optimizer = config_optimizer("adam"),
      validation = 0.2,
      device = "cuda",
      verbose = F,
      plot = F
    )
    
    save(nn_fit, file = paste0("data/r_objects/ssdm_results/full_models/", species, "_nn_fit.RData"))
  })
}
